Name:		whee
Version:	{{ version }}
Release:	1%{?dist}
Summary:	A wee set of files

Group:		Applications
License:	BSD-2-clause
URL:		https://www.debian.org/
Source0:	whee-{{ version }}.tar.gz

BuildArch:	noarch

# If we were not building this on a Debian system, we would also have this:
#BuildRequires:	rsync

%description
This is a package that may help with something.
Or not.

%prep
%autosetup -n whee-{{ version }}

%build
true

%install
rsync -a ./ '%{buildroot}/'
find '%{buildroot}/' -type f -exec touch -r /etc/passwd '{}' +

%files
/bin/*.bin
/etc/*.conf
/usr/share/whee/*.txt

%changelog
* Fri Dec 24 2021 Peter Pentchev <roam@ringlet.net> - {{ version }}-1
- Initial packaging; that's all there is.
